/*
 *-----------------------------------------------------------------------------
 * The confidential and proprietary information contained in this file may
 * only be used by a person authorised under and to the extent permitted
 * by a subsisting licensing agreement from Arm Limited or its affiliates.
 *
 *            (C) COPYRIGHT 2010-2013 Arm Limited or its affiliates.
 *                ALL RIGHTS RESERVED
 *
 * This entire notice must be reproduced on all copies of this file
 * and copies of this file may only be made by a person if such person is
 * permitted to do so under the terms of a subsisting license agreement
 * from Arm Limited or its affiliates.
 *
 *      SVN Information
 *
 *      Checked In          : $Date: 2017-10-10 15:55:38 +0100 (Tue, 10 Oct 2017) $
 *
 *      Revision            : $Revision: 371321 $
 *
 *      Release Information : Cortex-M System Design Kit-r1p1-00rel0
 *-----------------------------------------------------------------------------
 */

//==============================================================================
//
// CORTEX-M0+ CoreSight MTB Integration Test
//
//==============================================================================
//
// Simple example test case to showcase operation of CoreSight MTB-M0+
//
// Key test points included:
//      -  trace RAM address width setting
//      -  MTB pointer width setup
//      -  memory map of trace SRAM
//
// If MTB is available, then
//      1. Check MTB CoreSight Registers
//      2. Check integration of MTB SRAM
//      3. Trace simple code and check trace
//      4. Trace exception and check trace
//      5. Trace simple code using TSTART/STOP and check trace
//      6. Check MTB User config
//
// NOTE: Test relies on compiler not optimizing away empty functions
//

#ifdef CORTEX_M0
#include "CMSDK_CM0.h"
#include "core_cm0.h"
#else
#ifdef CORTEX_M0PLUS
#include "CMSDK_CM0plus.h"
#include "core_cm0plus.h"
#endif
#endif

#include <stdio.h>
#include "uart_stdout.h"
#include "CMSDK_driver.h"

#include "config_id.h"

//=====================================================================
// Function prototype
//=====================================================================
// This function reads the data from the MTB RAM.
uint32_t ReadMemMTB(uint32_t);
// This function includes 2 calls to dummy_func which translates to:
//
//    dummy_func  0x000001c8:    4770        pG      BX       lr
//
//    func_wrapper
//                0x000001ca:    b500        ..      PUSH     {lr}
//                0x000001cc:    f7fffffc    ....    BL       dummy_func ; 0x1c8
//                0x000001d0:    f7fffffa    ....    BL       dummy_func ; 0x1c8
//                0x000001d4:    bd00        ..      POP      {pc}
//
void func_wrapper(void);
// This function is provided so that a branch happens to it and BX lr
// is automatically generated when its exited even though there are no
// instructions present.
void dummy_func(void);
// SVC Handler function which does not do anything
void SVC_Handler(void);
// This function tests the results of the traced results for function
// func_wrapper (as descrivbed above).
void check_func_wrapper_trace(unsigned int trace_addr_offset);
// Checks that the MTB actual address width matches expected, makes
// sure that aliasing will not happen and that AHB address decode
// logic is sufficient to cover MTB's SRAM memory.
void check_mtb_awidth(void);
// Function ends test with fail appropriate message
void fail_terminate_test(void);
// This function checks the expected trace (3rd argument) value against
// the actual trace value (2nd argument) and prints fail message the
// string provided in argument 1.
void CheckHex(char *, uint32_t, uint32_t);
// Checks that the MTB is aligned correctly for the size address width
// versus base address.
void check_mtb_alignment(uint32_t, uint32_t);
// Reverses data for big endian.
uint32_t SwapEndian(uint32_t);

//=====================================================================
// Global Variables to track test progress
//=====================================================================
uint32_t *ptr;
uint32_t Errors = 0;        // error counter
uint32_t probing_mtb = 0;   // Indicate probing of MTB registers in progress. Use by HardFault handler

//
// Pointer to functions that provide the MTB trace with the branch into/out
// of function as well as entry to SVC handler.
//

#ifdef CORTEX_M0PLUS
void (*func_wrapper_ptr)(void) = func_wrapper;
void (*dummy_func_ptr)(void)   = dummy_func;
void (*svc_ptr)(void)          = SVC_Handler;
#endif

#if   defined ( __CC_ARM )
__svc(0) void my_svc(int); // this is for ARM C compiler or Keil MDK.
#endif

//=====================================================================
// Start of main program
//=====================================================================

int main (void)
{
#ifdef CORTEX_M0PLUS
  uint32_t svc_addr;           // address of SVC handler
  uint32_t expected_addr = 0;  // expected address variable
  uint32_t addr_var;           // address variable in checking loop
  uint32_t test_addr;          // trace data starting address to be used
  uint32_t end_addr;           // end of trace address
  uint32_t value;              // read value for mtb trace data check
  int      loop_counter;       // loop counter

  int need_to_wrap = 0;       // Whether the trace data checking needs to wrap
  int wrap         = 0;       // Set when there's a need to wrap, and cleared
                              //  when the wrap occurs
#endif

  // UART init
  UartStdOutInit();

#ifdef CORTEX_M0PLUS
  // Banner
  puts("CORTEX-M0+ - MTB Test\n");


  if (EXPECTED_MTB == 1) // This is defined in config_id.h
    {

      //--------------------------------------
      // Part 1. Check MTB CoreSight register
      //--------------------------------------

      // Probe MTB register
      probing_mtb = 1; // Indicate to HardFault handler (if triggered) it is probing MTB ID registers

      // These ID constants are defined in CMSDK_CM0plus.h

      puts("\nChecking MTB CoreSight registers:\n");
      CheckHex("MTB BASE", SwapEndian(MTB->BASE), CMSDK_MTBSRAM_BASE);
      CheckHex("MTB ITCTRL", SwapEndian(MTB->ITCTRL), 0x0);
      CheckHex("MTB CLAIMSET", SwapEndian(MTB->CLAIMSET), 0x0);
      CheckHex("MTB CLAIMCLR", SwapEndian(MTB->CLAIMCLR), 0x0);
      CheckHex("MTB LOCKACCESS", SwapEndian(MTB->LOCKACCESS), 0x0);
      CheckHex("MTB LOCKSTATUS", SwapEndian(MTB->LOCKSTATUS), 0x0);
      CheckHex("MTB AUTHSTATUS", SwapEndian(MTB->AUTHSTATUS), 0xF);
      CheckHex("MTB DEVARCH", SwapEndian(MTB->DEVARCH), 0x47700A31);
      CheckHex("MTB DEVID", SwapEndian(MTB->DEVID), 0x0);
      CheckHex("MTB DEVTYPE", SwapEndian(MTB->DEVTYPE), 0x31);
      CheckHex("MTB CID0", SwapEndian(MTB->CID0), 0xD);
      CheckHex("MTB CID1", SwapEndian(MTB->CID1), 0x90);
      CheckHex("MTB CID2", SwapEndian(MTB->CID2), 0x5);
      CheckHex("MTB CID3", SwapEndian(MTB->CID3), 0xB1);
      CheckHex("MTB PID0", SwapEndian(MTB->PID0), (CM0P_MTB_PART & 0xFF));
      CheckHex("MTB PID1", SwapEndian(MTB->PID1), ((ARM_JEP_ID<<4 | CM0P_MTB_PART>>8) & 0xFF));
      CheckHex("MTB PID2", SwapEndian(MTB->PID2), (CM0P_MTB_REV<<4 | 8 | ARM_JEP_ID>>4));
      CheckHex("MTB PID3", SwapEndian(MTB->PID3), CM0P_MTB_ECO);
      CheckHex("MTB PID4", SwapEndian(MTB->PID4), ARM_JEP_CONT);
      CheckHex("MTB PID5", SwapEndian(MTB->PID5), 0x0);
      CheckHex("MTB PID6", SwapEndian(MTB->PID6), 0x0);
      CheckHex("MTB PID7", SwapEndian(MTB->PID7), 0x0);

      probing_mtb = 0; // Test of accessing MTB registers is done.

      //--------------------------------------
      // Part 2. Check integration of MTB SRAM
      //--------------------------------------
      check_mtb_awidth();

      // Alignment check of base address vs width
      check_mtb_alignment(CMSDK_MTBSRAM_BASE, EXPECTED_MTB_AWIDTH);

      //--------------------------------------
      // Part 3. MTB Trace generation and check
      //--------------------------------------

      puts("\nPart 3: Tracing simple code\n");

      // POSITION POINTER = 0xf00, WRAP = 0,
      // 48 bytes of trace data space is needed for this test part
      // It can be moved to other address if needed
      test_addr     = 0xf00UL;  // need to save this for checking function
      MTB->POSITION = SwapEndian(test_addr);

      // Program FLOW to stop tracing at last address before wrapping
      MTB->FLOW = SwapEndian(((0x1UL << EXPECTED_MTB_AWIDTH)-1) & MTB_FLOW_WATERMARK_Msk);
      // Set MASTER MASK to maximum valid size = AWIDTH-4, which gives a size
      // of 2^AWIDTH bytes
      MTB->MASTER = SwapEndian((EXPECTED_MTB_AWIDTH-4) << MTB_MASTER_MASK_Pos);

      // Enable Trace
      MTB->MASTER = SwapEndian(SwapEndian(MTB->MASTER) | (0x1UL << MTB_MASTER_EN_Pos));

      // Code to trace
      // Call via the pointer to the function to ensure that the
      // compiler/linker can't perform any optimisations that would lead to the
      // trace data not being as expected
      // Initialise the function pointers here; because the test data and trace
      // data are being written to the same memory, there's a risk that they
      // will be written to the same memory addresses, so these pointers may
      // have been overwritten since they were first initialised, so
      // reinitialise them
      func_wrapper_ptr = func_wrapper;
      dummy_func_ptr   = dummy_func;
      func_wrapper_ptr();

      // Disable Trace
      MTB->MASTER = 0x0UL;

      // Check Trace result
      check_func_wrapper_trace(test_addr); // must match MTB->POSITION setting

      //--------------------------------------
      // Part 4. Trace exception and check
      //--------------------------------------

      puts("\nPart 4: Tracing exception\n");

      // POSITION POINTER = 0xf00, WRAP = 0,
      // 24 bytes of trace data space is needed for this part test.
      // It can be moved to other address if needed
      test_addr     = 0xf00UL;  // need to save this for checking function
      MTB->POSITION = SwapEndian(test_addr);

      // Set MASTER MASK to maximum valid size = AWIDTH-4, which gives a size
      // of 2^AWIDTH bytes
      MTB->MASTER = SwapEndian((EXPECTED_MTB_AWIDTH-4) << MTB_MASTER_MASK_Pos);

      // Enable Trace
      MTB->MASTER = SwapEndian(SwapEndian(MTB->MASTER) | (0x1UL << MTB_MASTER_EN_Pos));

      // Exception to trace

#if   defined ( __CC_ARM )
      my_svc(0);             // for Keil MDK-ARM or ARM C compiler
#else
      __asm("   svc  #0\n"); // gcc style
#endif

      // Disable Trace
      MTB->MASTER = 0x0UL;

      // Checking MTB Trace
      //
      // The main thread executes a SVC
      // 3 branches are expected. First to SVC Handler and
      // 2 on exception return of which one is to
      // EXC_RETURN value. The source address of
      // the first branch and the destination address of
      // the last branch can't be reliably checked
      //
      // extract SVC Handler address from global function pointer
      // Initialise the handler pointer here; because the test data and trace
      // data are being written to the same memory, there's a risk that they
      // will be written to the same memory addresses, so this pointer may have
      // been overwritten since it was first initialised, so reinitialise it
      svc_ptr      = SVC_Handler;
      svc_addr     = (((uint32_t)svc_ptr) & ~0x1UL); // clear LSB which is 1 to indicate Thumb state
      addr_var     = CMSDK_MTBSRAM_BASE+test_addr;   // starting address for trace data extraction
      loop_counter = 1;                              // starting with first packet

      // Calculate ending address
      end_addr = (CMSDK_MTBSRAM_BASE + (SwapEndian(MTB->POSITION) & 0xFFFFFFF8));

      // It is possible that the trace data wrapped to the bottom of memory
      // depending on the address width of the MTB SRAM; so the checking loop
      // needs to take this into account
      // If the end address is less than the starting address, then calculate the
      // number of entries from start address to top of mem, and from bottom of mem
      // to end address
      if (end_addr < addr_var) {
        // If the starting address is off the top of the memory though, the only
        // entries will be from the bottom of memory to the and address
        if (addr_var < (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH))) {
          // Record the fact that the trace data checking algorithm will need to wrap
          need_to_wrap = 1;
        } else {
          // The start address is off the top of memory, so the actual start
          // address will be the bottom of memory
          // Set addr_var to be the bottom of memory
          addr_var = CMSDK_MTBSRAM_BASE;
        }
      }
      // Set wrap if there's a need to wrap; wrap will be cleared once the wrap has
      // occurred so that the loop termination is then controlled by the current
      // address being checked meets the end address
      wrap = need_to_wrap;
      while ( // while we haven't wrapped, and the current address is less than end of mem
              (1==wrap && (addr_var < (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH)))) ||
              // or we have wrapped, or there was no need to wrap, and the current
              // address is less than the end address
              (addr_var < end_addr))
      {
        // Print format- [trace packet number since last enable] : D(Destination
        //              address) %x S(Source address) %x
        printf("[%d] : D 0x%x, S 0x%x\n",loop_counter,ReadMemMTB(addr_var+0x4UL),ReadMemMTB(addr_var));

        switch(loop_counter)
        {

      // 3 packets:                      |      source      |        dest       |  Note:
      //   - SVC exception triggered     |                  |    svc_handler    | atomic bit = 1
      //   - SVC exception exit          |   svc_handler    |    exc_return     | 0xFFFFFFF9 in this case
      //   - 1st dummy_func returns      |   exc_return     |                   | atomic bit = 1

          case 1:
            // Check Atomic bit and destination address
            value = ReadMemMTB(addr_var);
            CheckHex("- Trace packet 1, Atomic bit", (value & 0x1), 0x1);
            value = ReadMemMTB(addr_var+0x4UL);
            expected_addr = svc_addr;
            CheckHex("- Trace packet 1, Dest addr", (value & ~0x1UL), expected_addr);
            break;

          case 2:
            // Check Atomic bit and source address
            value = ReadMemMTB(addr_var);
            CheckHex("- Trace packet 2, Atomic bit", (value & 0x1), 0x0);
            expected_addr = svc_addr;
            CheckHex("- Trace packet 2, Src  addr", (value & ~0x1UL), expected_addr);
            // Check destination address
            value = ReadMemMTB(addr_var+0x4UL);
            expected_addr = (0xFFFFFFF9 & ~0x1UL);
            CheckHex("- Trace packet 2, Dest addr", (value & ~0x1UL), expected_addr);
            break;

          case 3:
            // Check Atomic bit and source address
            value = ReadMemMTB(addr_var);
            CheckHex("- Trace packet 3, Atomic bit", (value & 0x1), 0x1);
            expected_addr = (0xFFFFFFF9 & ~0x1UL);
            CheckHex("- Trace packet 3, Src  addr", (value & ~0x1UL), expected_addr);
            break;

          default: // More trace packet than expected
            puts("Unexpected Trace packet\n");
            Errors++;
        }

        addr_var += 0x8; // Each packet is 8 bytes in size
        loop_counter++;

        // Wrap the address if necessary
        if (addr_var >= (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH))) {
          addr_var = CMSDK_MTBSRAM_BASE;
          // Clear wrap now that we've wrapped
          wrap = 0;
        }
      }

      //-------------------------------------------------------
      // Part 5. Trace simple code using TSTART/STOP and check
      //-------------------------------------------------------

      puts("\nPart 5: Tracing using TSTART and TSTOP\n");

      // Trace start   TSTART = P1[7]
      // Trace stop    TSTOP  = P1[6]
      CMSDK_GPIO1->LB_MASKED[((1<<6) | (1<<7))] = 0; // Set the two output pins as zero
      CMSDK_GPIO1->OUTENABLESET = (1<<6) | (1<<7);   // Set these two pins as output

      // POSITION POINTER = 0xf00, WRAP = 0
      // 48 bytes of trace data space is needed for this part test.
      // It can be moved to other address if needed, but if doing so, also need to update
      //   the input parameter for check_func_wrapper_trace().
      test_addr     = 0xf00UL;
      MTB->POSITION = SwapEndian(test_addr);

      // Program FLOW to stop tracing at last address before wrapping
      MTB->FLOW = SwapEndian(((0x1UL << EXPECTED_MTB_AWIDTH)-1) & MTB_FLOW_WATERMARK_Msk);
      // Set MASTER MASK = 16-4 (Size = 64KB)
      MTB->MASTER = SwapEndian(((EXPECTED_MTB_AWIDTH-4) << MTB_MASTER_MASK_Pos) |
                               (0x1UL << MTB_MASTER_TSTARTEN_Pos) |
                               (0x1UL << MTB_MASTER_TSTOPEN_Pos));

      // Generate a pulse on TSTART
      CMSDK_GPIO1->LB_MASKED[((1<<7))] = (1<<7); // Set   TSTART
      CMSDK_GPIO1->LB_MASKED[((1<<7))] = 0; // Clear TSTART

      __NOP(); // delay to ensure TSTART propagated
      __NOP();
      __NOP();
      __NOP();
      __NOP();


      // Code to trace
      // Call via the pointer to the function to ensure that the
      // compiler/linker can't perform any optimisations that would lead to the
      // trace data not being as expected
      // Initialise the function pointers here; because the test data and trace
      // data are being written to the same memory, there's a risk that they
      // will be written to the same memory addresses, so these pointers may
      // have been overwritten since they were first initialised, so
      // reinitialise them
      func_wrapper_ptr = func_wrapper;
      dummy_func_ptr   = dummy_func;
      func_wrapper_ptr();

      // Generate a pulse on TSTOP
      CMSDK_GPIO1->LB_MASKED[((1<<6))] = (1<<6); // Set TSTOP
      CMSDK_GPIO1->LB_MASKED[((1<<6))] = 0; // Clear TSTOP

      __NOP(); // delay to ensure TSTOP propagated
      __NOP();
      __NOP();
      __NOP();
      __NOP();

      // Calculate ending address
      end_addr = (CMSDK_MTBSRAM_BASE + (SwapEndian(MTB->POSITION) & 0xFFFFFFF8));

      //printf ("- Start addr: 0x%X\n", (CMSDK_MTBSRAM_BASE+test_addr));
      //printf ("- Stop  addr: 0x%X\n", end_addr);

      if ((CMSDK_MTBSRAM_BASE+test_addr)==end_addr) {
        puts ("ERROR : no trace detected. TSTART / TSTOP might not be connected correctly\n");
        Errors++;
	}
      else {
        // Check Trace result
        check_func_wrapper_trace(test_addr);
      }
      //--------------------------------------
      // Part 6. Check MTB User configuration
      //--------------------------------------

      puts("\nPart 6: User configuration\n");

      // Set only the SFRWPRIV and RAMPRIV bits in MTB MASTER register
      MTB->MASTER = SwapEndian( MTB_MASTER_RAMPRIV_Msk | MTB_MASTER_SFRWPRIV_Msk );

      // Read back MTB MASTER register
      value = SwapEndian(MTB->MASTER);

      // Check if both bits were set
      value = ((value & MTB_MASTER_RAMPRIV_Msk)  == MTB_MASTER_RAMPRIV_Msk) &
              ((value & MTB_MASTER_SFRWPRIV_Msk) == MTB_MASTER_SFRWPRIV_Msk);

      // value   EXPECTED_USER
      //   0          0             Correct (User mode not available)
      //   1          1             Correct (User mode available)
      //   0          1             Incorrect (configuration mismatched)
      //   1          0             Incorrect (configuration mismatched)

      if(value == EXPECTED_USER)
        {
          puts("MTB User configuration check\t-\tPASS\n");
        }
      else
        {
          puts("MTB User configuration check\t-\tFAIL\n");
	  puts("- Please check the setting of EXPECTED_USER setting");
          Errors++;
        }

      // Set only the SFRWPRIV bit in MTB MASTER register since RAM is accessed
      // for data as well as for Trace
      MTB->MASTER = SwapEndian(MTB_MASTER_SFRWPRIV_Msk);

      // if USER mode exists try clearing MASTER register in USER mode
      if(EXPECTED_USER)
        {
	  // Check EXPECTED_BE is set correctly
          if ( ((EXPECTED_BE) && ((SCB->AIRCR & SCB_AIRCR_ENDIANESS_Msk)==0) )||
	       ((EXPECTED_BE==0) && ((SCB->AIRCR & SCB_AIRCR_ENDIANESS_Msk)!=0)) )
	    {
              puts("ERROR: EXPECTED_BE setting does not match detected setting\n");
              Errors++;
	    }
          // Change main thread to USER mode
          __set_CONTROL(1);

          // Try to clear MTB MASTER register
          MTB->MASTER = 0;

         // Read back MTB MASTER register
          value = MTB->MASTER;

          // Account for Endianness
	  // Note: Now it is in user mode, therefore cannot access to SCB registers
	  //       so we rely on EXPECTED_BE setting
          if (EXPECTED_BE) value = __REV(value);

          // Check if SFR PRIV bit is still set as user mode write has no effect
          if(((value & MTB_MASTER_SFRWPRIV_Msk) != 0))
            { // HPROT[1] indicate an access is in user access level
              puts("MTB HPROT[1] integration check\t-\tPASS\n");
            }
          else
            {
              puts("MTB HPROT[1] integration check\t-\tFAIL\n");
              Errors++;
            }
        }

    }
    else
      { // Checking if the system could support MTB but EXPECTED_MTB is not setup correctly.
        probing_mtb = 2;
        puts("	EXPECTED_MTB is set to 0 (not present)\n");
	puts("	Access to MTB register should result in HardFault.\n");

        CheckHex("MTB DEVARCH", SwapEndian(MTB->DEVARCH), 0x47700A31);

	// Should not be here : No fault occurred, that's not expected.
        puts("ERROR : MTB ID found\n");
	Errors++;
      }

  // Print final test result
  if (Errors == 0) {
     puts("** TEST PASSED ** \n");
  } else {
     printf("** TEST FAILED ** with %d errors\n", Errors);
  }

#else
  puts("** TEST SKIPPED ** CORTEX-M0 does not support MTB\n");
#endif

  UartEndSimulation();  //stop simulation

  return Errors;
} // End of main program


#ifdef CORTEX_M0PLUS
// ----------------------------------------------------------------------------
//
// Functions being traced
//
// ----------------------------------------------------------------------------
void func_wrapper(void)
{
    // Call via the pointer to the function to ensure that the compiler/linker
    // can't perform any optimisations that would lead to the trace data not
    // being as expected
    dummy_func_ptr();
    dummy_func_ptr();
}

void dummy_func(void)
{
// Does nothing
}

void SVC_Handler(void)
{
// Does nothing
}


// ----------------------------------------------------------------------------
// Trace result checking for part 3 and part 5
// ----------------------------------------------------------------------------
void check_func_wrapper_trace(unsigned int trace_addr_offset)
{
  // Checking MTB Trace
  //
  // The main thread calls func_wrapper
  // func_wrapper in turn calls empty function dummy_func 2 times
  // This results in 6 branches assuming the compiler hasn't blown
  // away empty functions. Of the 6 branches the source address of
  // the first branch and the destination address of the last
  // branch can't be reliably checked
  //
  // Assembly code expected for function calls is
  // dummy_func
  //            BX lr           (Thumb)
  // func_wrapper
  //            PUSH {lr}       (Thumb)
  //            BL dummy_func   (Thumb2)
  //            BL dummy_func   (Thumb2)
  //            POP {pc}        (Thumb)
  //
  uint32_t func_wrapper_addr; // extracted function address for func_wrapper()
  uint32_t dummy_func_addr;   // extracted function address for dummy_func()
  uint32_t expected_addr;     // expected address variable
  uint32_t addr_var;          // address variable
  uint32_t end_addr;          // end of trace address
  uint32_t value;             // read value for mtb trace data check
  uint32_t entries;           // the number of entries in the MTB
  int      loop_counter;      // loop counter

  int found_expected;

  int need_to_wrap = 0;       // Whether the trace data checking needs to wrap
  int wrap         = 0;       // Set when there's a need to wrap, and cleared
                              //  when the wrap occurs

  // extract function addresses from global function pointers that are already
  // initialised
  func_wrapper_addr = (((uint32_t)func_wrapper_ptr) & ~0x1UL);
  dummy_func_addr   = (((uint32_t)dummy_func_ptr) & ~0x1UL);

  addr_var = CMSDK_MTBSRAM_BASE+trace_addr_offset;
  loop_counter = 1;

  if (EXPECTED_MTB_AWIDTH < 6)
    {
      puts("WARNING: Not enough MTB SRAM to trace all packets. Hence cannot check all packets!\n");
    }

  if(SwapEndian(MTB->POSITION) == 0x0UL)
    {
       puts("Trace expected, none found -\tFAIL\n");
       Errors++;
    }

  // Calculate ending address
  end_addr = (CMSDK_MTBSRAM_BASE + (SwapEndian(MTB->POSITION) & 0xFFFFFFF8));

  // Check that number of packets is a sensible number for the expected sequence
  // Note that as a sequence of C code is being traced, the exact number of
  // branches cannot be relied upon (for example, the compiler may insert
  // string literals and branch over them, hence inserting extra branch
  // instructions)
  // The MTB stores pairs of addresses (source and destination), so each
  // 'entry' is 8 bytes; convert from bytes to entries by dividing by 8 therefore
  // It is possible that the trace data wrapped to the bottom of memory
  // depending on the address width of the MTB SRAM; so the number of entries
  // calculation needs to take this into account
  // If the end address is less than the starting address, then calculate the
  // number of entries from start address to top of mem, and from bottom of mem
  // to end address
  if (end_addr < addr_var) {
    // If the starting address is off the top of the memory though, the only
    // entries will be from the bottom of memory to the and address
    if (addr_var < (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH))) {
      // The start address is less than top of mem, so calculate entries from
      // start address to top of mem
      entries  = ((CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH)) - addr_var) >> 3;
      // Record the fact that the trace data checking algorithm will need to wrap
      need_to_wrap = 1;
    } else {
      // The start address is off the top of memory, so the actual start
      // address will be the bottom of memory, so the number of entries from
      // start to the top of memory is zero
      entries  = 0;
      // Set addr_var to be the bottom of memory
      addr_var = CMSDK_MTBSRAM_BASE;
    }
    // Add in the number of entries from bottom of mem to end address
    entries += (end_addr - CMSDK_MTBSRAM_BASE) >> 3;
  } else {
    // Otherwise it's the normal case of no wrapping having occurred, so
    // calculate the number of entries from start address to end address
    entries = (end_addr - addr_var) >> 3;
  }
  if (entries > 15) {
    printf("Excessive number of branches traced (%d); expecting 6, but allowing 15\n", entries);
    Errors++;
  }

  // First: find the first expected packet - we allow other packets to be
  // present before the expected, since with compiled code, the exact number of
  // branches cannot be relied upon
  found_expected = 0;
  // Set wrap if there's a need to wrap; wrap will be cleared once the wrap has
  // occurred so that the loop termination is then controlled by the current
  // address being checked meets the end address
  wrap = need_to_wrap;
  while (( // while we haven't wrapped, and the current address is less than end of mem
           (1==wrap && (addr_var < (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH)))) ||
           // or we have wrapped, or there was no need to wrap, and the current
           // address is less than the end address
           (addr_var < end_addr)
           // and the expected starting trace packet hasn't been found
          ) && (found_expected != 1)) {
    // Read the value from memory
    value = ReadMemMTB(addr_var+0x4UL);
    expected_addr = func_wrapper_addr;
    // remove bit 0 of the value, since this is the start bit
    value &= ~0x1UL;
    // if this is the expected address, set found_expected
    // otherwise increment the addr, and continue looking
    if (value == expected_addr) {
      found_expected = 1;
    } else {
      addr_var += 0x8; // Each packet is 8 bytes in size
      // Wrap the address if necessary
      if (addr_var >= (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH))) {
        addr_var = CMSDK_MTBSRAM_BASE;
        // Clear wrap now that we've wrapped
        wrap = 0;
      }
    }
  }

  // Set wrap if there's a need to wrap; wrap will be cleared once the wrap has
  // occurred so that the loop termination is then controlled by the current
  // address being checked meets the end address
  wrap = need_to_wrap;
  // Read trace packets until end of trace
  while ( // while we haven't wrapped, and the current address is less than end of mem
          (1==wrap && (addr_var < (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH)))) ||
          // or we have wrapped, or there was no need to wrap, and the current
          // address is less than the end address
          (addr_var < end_addr))
  {
    // Print format- [trace packet number since last enable] : D(Destination
    //              address) %x S(Source address) %x
    printf("[%d] : D 0x%x, S 0x%x\n",loop_counter,ReadMemMTB(addr_var+0x4UL),ReadMemMTB(addr_var));

    switch(loop_counter)
    {
      // 6 packets:                      |      source      |        dest       |  Note:
      //   - branch to func_wrapper      | not determinable |    func_wrapper   |
      //   - 1st branch to dummy_func    |   func_wrapper+2 |    dummy_func     | (PUSH is 2 bytes)
      //   - 1st dummy_func returns      |   dummy_func     |  func_wrapper+2+4 | (BL is 4 bytes)
      //   - 2nd branch to dummy_func    | func_wrapper+2+4 |    dummy_func     |
      //   - 2nd dummy_func returns      |   dummy_func     |func_wrapper+2+4+4 |
      //   - func_wrapper returns        |func_wrapper+2+4+4| not determinable  |
      // Atomic bits are all 0.
      // Some of the expected addresses are not deterministic, and depend on
      // what code the compiler has generated, so some of the above values need
      // to be ignored:
      //  - first branch to dummy_func - source may vary, so ignore;
      //    destination is fixed at dummy_func
      //  - first return from dummy_func - destination may vary, so ignore;
      //    source is fixed at dummy_func (single instruction BX lr)
      //  - second branch to dummy_func - source may vary, as above
      //  - second return from dummy func - destination may vary, as above
      //  - return from func_wrapper - source may vary, so ignore

      case 1:
        // Comments from above:
        //   - branch to func_wrapper      | not determinable |    func_wrapper   |
        //   - source address is not determinable
        //
        // Check Atomic bit (source address cannot be easily checked)
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 1, Atomic bit", (value & 0x1), 0x0);
        // Check start bit and destination address
        value = ReadMemMTB(addr_var+0x4UL);
        CheckHex("- Trace packet 1, Start bit", (value & 0x1), 0x1);
        expected_addr = func_wrapper_addr;
        CheckHex("- Trace packet 1, Dest addr", (value & ~0x1UL), expected_addr);
        break;

      case 2:
        // Comments from above:
        //   - 1st branch to dummy_func    |   func_wrapper+2 |    dummy_func     | (PUSH is 2 bytes)
        //   - first branch to dummy_func - source may vary, so ignore;
        //    destination is fixed at dummy_func
        //
        // Check Atomic bit and source address
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 2, Atomic bit", (value & 0x1), 0x0);
        // Ignore source address, as described above
        // Check destination address
        value = ReadMemMTB(addr_var+0x4UL);
        expected_addr = dummy_func_addr;
        CheckHex("- Trace packet 2, Dest addr", (value & ~0x1UL), expected_addr);
        break;

      case 3:
        // Comments from above:
        //   - 1st dummy_func returns      |   dummy_func     |  func_wrapper+2+4 | (BL is 4 bytes)
        //   - first return from dummy_func - destination may vary, so ignore;
        //    source is fixed at dummy_func (single instruction BX lr)
        //
        // Check Atomic bit and source address
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 3, Atomic bit", (value & 0x1), 0x0);
        // Check source address
        expected_addr = dummy_func_addr;
        CheckHex("- Trace packet 3, Src  addr", (value & ~0x1UL), expected_addr);
        // Ignore destination address, as described above
        break;

      case 4:
        // Comments from above:
        //   - 2nd branch to dummy_func    | func_wrapper+2+4 |    dummy_func     |
        //   - second branch to dummy_func - source may vary, as above
        //
        // Check Atomic bit and source address
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 4, Atomic bit", (value & 0x1), 0x0);
        // Ignore source address, as described above
        // Check destination address
        value = ReadMemMTB(addr_var+0x4UL);
        expected_addr = dummy_func_addr;
        CheckHex("- Trace packet 4, Dest addr", (value & ~0x1UL), expected_addr);
        break;

      case 5:
        // Comments from above:
        //   - 2nd dummy_func returns      |   dummy_func     |func_wrapper+2+4+4 |
        //   - second return from dummy func - destination may vary, as above
        //
        // Check Atomic bit and source address
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 5, Atomic bit", (value & 0x1), 0x0);
        // Check source address
        expected_addr = dummy_func_addr;
        CheckHex("- Trace packet 5, Src  addr", (value & ~0x1UL), expected_addr);
        // Ignore destination address, as described above
        break;

      case 6:
        // Comments from above:
        //   - func_wrapper returns        |func_wrapper+2+4+4| not determinable  |
        //   - return from func_wrapper - source may vary, so ignore
        //     destination is not determinable
        //
        // Check Atomic bit and source address
        value = ReadMemMTB(addr_var);
        CheckHex("- Trace packet 6, Atomic bit", (value & 0x1), 0x0);
        // Ignore source address, as described above
        // Ignore destination address, as described above
        break;

      // There is a check above for there being more trace than expected; this
      // check cannot be precise, as explained above
      // So there is no requirement to perform the check here
     }

    addr_var += 0x8; // Each packet is 8 bytes in size
    loop_counter++;

    // Wrap the address if necessary
    if (addr_var >= (CMSDK_MTBSRAM_BASE + (1<<EXPECTED_MTB_AWIDTH))) {
      addr_var = CMSDK_MTBSRAM_BASE;
      // Clear wrap now that we've wrapped
      wrap = 0;
    }

  } // end while_loop

} // end of checking (check_func_wrapper_trace)


// ----------------------------------------------------------------------------
// Test function to check if the address width of MTB and MTB memory are
// setup correctly
// ----------------------------------------------------------------------------

void check_mtb_awidth(void)
{
  // The purpose of this function are the following:
  // 2a. Check that EXPECTED_MTB_AWIDTH and the actual address width
  //     of the MTB match.
  // 2b. Check that MTB SRAM's address width is not less than MTB's
  //     address width. Smaller SRAM size will result in aliasing.
  // 2c. Check that AHB address decode logic in the example system
  //     provide MTB memory map address range is sufficient to cover
  //     MTB's SRAM memory.
  //
  // Part 2a. is tested by writing 1's to MTB's POINTER and reading
  // back. If they don't match, the test is terminated.
  //  //
  // Part 2b. and 2c. are tested by first making the CPU to initialize
  // the memory starting from the top and moving down in step-sizes.
  // The memory-under-test is top-most address. The MTB traces a value
  // different from the initialized value into the memory-under-test.
  // This traced value is read back and checked.
  // - If it is the traced address value, the MTB can access this memory and
  //    so can the CPU.
  // - If the value is neither the traced value nor the initialized value, it
  //    means the CPU doesn't have visibility of this memory location,
  //    the test is then terminated.
  //
  // It is also possible that the CPU access to the MTB memory space can fault
  // because the system's AHB interconnect maps this region to other devices say
  // the AHB Default Slave. If this is the case, the test will take a hard fault
  // and terminate.
  //
  // If the above check on the memory-under-test passed, the test checks
  // if the lower memory locations (moving down by the step-size)
  // still has the initialization value. If any of the lower addresses has
  // the traced address value, it means aliasing has happened.
  // If such aliasing problem is detected, the test will terminate as this
  // means the SRAM size is smaller than MTB's address width.
  //
  // If there isn't any aliasing, the lower memory location should return
  // the initialized data value. If the initialized value is not seen, the
  // test terminates (memory corrupted for unknown reason).
  //
  // If the test hasn't terminated, then the test sequence is then repeated
  // with next memory-under-test becomes the current memory-under-test minus the
  // step-size. The checks are repeated until the lowest memory in the
  // loop. This lowest memory is not CMSDK_MTBSRAM_BASE. Hence a separate check
  // exists for CMSDK_MTBSRAM_BASE.
  //
  // The step-size here has been fixed at 512 bytes. This can be made smaller
  // but will slow down the algorithm. This algorithm is fast for reasonable
  // memory sizes.
  //
  // local variables
  uint32_t tmpvar = 0;         // temp variable
  uint32_t stepsize = 0x200UL; // step size for address in loops
  uint32_t addr_var = 0;       // address variable
  uint32_t expected_addr = 0;  // expected address variable

  puts("\nPart 2:MTB address width check\n");
  if(EXPECTED_MTB_AWIDTH > 2) // test requires minimum of 32 bytes
    {

      // PART 2A
      MTB->POSITION = SwapEndian(0xFFFFFFF8UL);

      MTB->FLOW     = SwapEndian(((0x1UL << EXPECTED_MTB_AWIDTH)-1) & MTB_FLOW_WATERMARK_Msk);

      // expected value of MTB->POSITION generated from EXPECTED_MTB_AWIDTH
      tmpvar = ((1UL << EXPECTED_MTB_AWIDTH) - 1) & 0xFFFFFFF8UL;

      if(tmpvar != SwapEndian(MTB->POSITION))
        {
          puts("Failed: check_mtb_awidth (part 2a)\n - Expected MTB Address width does not match actual\t-\tFAIL\n");
          fail_terminate_test();
        }
      else {
         puts("Passed :check_mtb_awidth (part 2a)\n");
	 }

      // PART 2B & 2C

      // Initializing memory from the CPU
      // The used address locations are spaced with a interval defined as stepsize.
      // The default settings :
      //   - EXPECTED_MTB_AWIDTH = 16 (64KB)
      //   - step size = 0x200 (512, can be adjusted dynamically to speed up the simulation)
      //   - number of steps = 65536/512 - 1 = 127
      // (the Initialization value is 0xA5A5A5A5)
      if (EXPECTED_MTB_AWIDTH>14) // (over 32KB)
        { // increase step size for larger memory
	  stepsize = 1<<(EXPECTED_MTB_AWIDTH-6);
	  printf (" stepsize adjusted to %d\n",stepsize );
	}
      /*

      CMSDK_MTBSRAM_BASE + size  |----------------------|
                                 |                      | <- addr = ((1UL << EXPECTED_MTB_AWIDTH) - 4UL) + CMSDK_MTBSRAM_BASE
                                 |                      |    Note: the 1st one (top) not used
                                 |     (stepsize)       |
                                 |                      |
                                 |     A5A5A5A5         | <- 1st location
                                 |                      |
                                 |     (stepsize)       |
                                 |                      |
                                 |     A5A5A5A5         | <- 2nd location
                                 ...
      CMSDK_MTBSRAM_BASE         |----------------------|

      */

      // Start from top address
      addr_var = ((1UL << EXPECTED_MTB_AWIDTH) - 4UL) + CMSDK_MTBSRAM_BASE;

      // First step not used:
      //   For the configuration where a single RAM is used the top end of
      //   memory for data and MTB is not used since data will be lost, e.g variables
      //   like "Errors" are overwritten since when the MTB wraps it updates this
      //   value.
      addr_var = addr_var - stepsize;

      while(addr_var > CMSDK_MTBSRAM_BASE)
        {
          ptr = (uint32_t*)addr_var;
          // Initialize memory
          *ptr = SwapEndian(0xA5A5A5A5);
          // puts("Initializing %x to 0xA5A5A5A5\n",addr_var);
          addr_var = addr_var - stepsize;
        }

      // Again start from top address
      addr_var = ((1UL << EXPECTED_MTB_AWIDTH) - 4UL) + CMSDK_MTBSRAM_BASE;

      // Again, First step not used:
      addr_var = addr_var - stepsize;

      // extract function address from global function pointer
      // Initialise the function pointer here; because the test data and trace
      // data are being written to the same memory, there's a risk that they
      // will be written to the same memory addresses, so this pointer may have
      // been overwritten since it was first initialised, so reinitialise it
      dummy_func_ptr = dummy_func;
      expected_addr = (((uint32_t)dummy_func_ptr) & ~0x1UL); // LSB need to be cleared for address value
      // Normally LSB of function pointers is set to 1 to indicate Thumb state.

      while(addr_var > CMSDK_MTBSRAM_BASE)
        {
          // Put trace in addr_var.
          // As POSITION is double-word aligned, the destination
          // address of first trace packet will be put in addr_var

          // POSITION POINTER = addr_var, WRAP = 0       (Address of trace buffer)
          MTB->POSITION = SwapEndian(addr_var);
          // Set MASTER MASK = EXPECTED_MTB_AWIDTH-4 (Size of trace buffer)
          MTB->MASTER = SwapEndian((EXPECTED_MTB_AWIDTH-4) << MTB_MASTER_MASK_Pos);
          // Enable Trace
          MTB->MASTER = SwapEndian(SwapEndian(MTB->MASTER) | (0x1UL << MTB_MASTER_EN_Pos));
          // A function call to the traced function
          // Call via the pointer to the function to ensure that the
          // compiler/linker can't perform any optimisations that would lead to
          // the trace data not being as expected
          dummy_func_ptr();
          // Disable Trace
          MTB->MASTER = 0x0UL;

          // Check addr
          if((ReadMemMTB(addr_var) & ~0x1UL) != expected_addr)
            {
              printf("<Err1> Expected %x at address %x\t-\tFAIL\n",expected_addr,addr_var);
              fail_terminate_test();
            }

//#define SHORTER_TEST
#ifdef  SHORTER_TEST
          // Check next lower addresses for aliasing or initialized value.
	  // the expected value is 0xA5A5A5A5
          tmpvar = addr_var - stepsize;
          if (tmpvar > CMSDK_MTBSRAM_BASE)
            { // If any one of the subseqence locations got the traced address
              // the memory mapping is incorrect as alias occurred.
              // the bit zero of the trace result has to be filtered out (atomic bit)
              if((ReadMemMTB(tmpvar) & ~0x1UL) == expected_addr)
                {
                  printf("%x has aliased to address %x\t-\tFAIL\n",expected_addr,tmpvar);
                  fail_terminate_test();
                }
              else if(ReadMemMTB(tmpvar) != 0xA5A5A5A5UL)
                { // the subsequent trace memory should be untouched, so they should still be 0xA5A5A5A5
                  printf("<Err2> Expected %x at address %x\t-\tFAIL\n",0xA5A5A5A5UL,addr_var);
                  fail_terminate_test();
                }
	    }
#else
          // Check lower addresses for aliasing or initialized value
          tmpvar = addr_var - stepsize;
          while(tmpvar > CMSDK_MTBSRAM_BASE)
            { // If any one of the subseqence locations got the traced address
              // the memory mapping is incorrect as alias occurred.
              // the bit zero of the trace result has to be filtered out (atomic bit)
              if((ReadMemMTB(tmpvar) & ~0x1UL) == expected_addr)
                {
                  printf("%x has aliased to address %x\t-\tFAIL\n",expected_addr,tmpvar);
                  fail_terminate_test();
                }
              else if(ReadMemMTB(tmpvar) != 0xA5A5A5A5UL)
                { // the subsequent trace memory should be untouched, so they should still be 0xA5A5A5A5
                  printf("<Err2> Expected %x at address %x\t-\tFAIL\n",0xA5A5A5A5UL,addr_var);
                  fail_terminate_test();
                }
              tmpvar = tmpvar - stepsize;
            } // end while (2nd level)
#endif
          addr_var = addr_var - stepsize;
        }

      puts("\nPassed :check_mtb_awidth (part 2b)\n");

      // Part 2c
      // Also trace to bottom address plus an offset and read back
      addr_var = CMSDK_MTBSRAM_BASE+0x80UL;

      // POSITION POINTER = addr_var-4, WRAP = 0
      // Placing the predictable address i.e. destination address in the bottom address
      MTB->POSITION = SwapEndian(addr_var-4UL);
      // Set MASTER MASK = EXPECTED_MTB_AWIDTH-4
      MTB->MASTER = SwapEndian((EXPECTED_MTB_AWIDTH-4) << MTB_MASTER_MASK_Pos);
      // Enable Trace
      MTB->MASTER = SwapEndian(SwapEndian(MTB->MASTER) | (0x1UL << MTB_MASTER_EN_Pos));
      // A function call to the function to be traced
      // Call via the pointer to the function to ensure that the
      // compiler/linker can't perform any optimisations that would lead to the
      // trace data not being as expected
      dummy_func_ptr();
      // Disable Trace
      MTB->MASTER = 0x0UL;

      // Check
      if((ReadMemMTB(addr_var) & ~0x1UL) != expected_addr)
        {
          printf("<Err3> Expected %x at address %x\t-\tFAIL\n",expected_addr,addr_var);
          fail_terminate_test();
        }
      puts("\nPassed :check_mtb_awidth (part 2c)\n");

      puts("Part2:MTB SRAM Integration\t-\tPASS\n");
    }
  else
    {
      puts("MTB RAM Address Width must be at least 3 bits\n");
      puts("- Check EXPECTED_MTB_AWIDTH variable\n");
      UartEndSimulation();
    }
}

// ----------------------------------------------------------------------------
// Helper function to read MTB trace data with endian handling:
//   Reads from address location passed to it via 'address' and reverses value
//   accordingly based on the endianess of the memory.
//
// ----------------------------------------------------------------------------
uint32_t ReadMemMTB(uint32_t address)
{
  __IO uint32_t val = 0;
  uint32_t *a = (uint32_t*) address;
  val = *a;

  return SwapEndian(val);
}

// ----------------------------------------------------------------------------
// Helper function to compare expected and actual values
// Display error message if mismatched is found.
// This function is typically used in validating trace packets.
// ----------------------------------------------------------------------------
//
void CheckHex(char *name, uint32_t actual, uint32_t expected)
{
  if(actual == expected)
    {
      printf("%s: 0x%x\t-\tPASS\n",
           name,
           actual);
    }
  else
    {
      printf("%s: Actual 0x%x\t(Expected 0x%x)\t-\tFAIL\n",
           name,
           actual,
           expected);
      Errors++;
    }
}

// ----------------------------------------------------------------------------
// Checks the alignment of the base address with the address width of the MTB.
// Displays error message if mismatched is found.
// ----------------------------------------------------------------------------
//
void check_mtb_alignment(uint32_t baseaddr, uint32_t mtb_a_width)
{

  int  shiftmtbleft, shiftmtbright;

  shiftmtbright = (baseaddr >> mtb_a_width);
  shiftmtbleft  = (shiftmtbright << mtb_a_width);

  if (shiftmtbleft == baseaddr)
    {
      puts("Passed :check_mtb_alignment (part 2d)\n");
    }
  else
    {
      puts("\n**FAILED ** MTB base address is not aligned to the address width\n");
      UartEndSimulation();  //stop simulation
    }

}

// ----------------------------------------------------------------------------
// Helper function to allow test to work in Big Endian system:
// Reverses the data when it is big endian otherwise leaves unchanged.
// (privileged access only, and only handle 32-bit data)
// ----------------------------------------------------------------------------
//
uint32_t SwapEndian(uint32_t data)
{
  if(SCB->AIRCR & SCB_AIRCR_ENDIANESS_Msk) // Check endian status
    {
      return __REV(data); // reverse bytes
    }
  else
    {
      return data;
    }
}

// ----------------------------------------------------------------------------
// HardFault handler
// ----------------------------------------------------------------------------
void HardFault_Handler()
{
  if (probing_mtb==1) {
    puts("[HardFault]");
    puts("- triggered when trying to access MTB register.");
    puts("- check EXPECTED_MTB in config_id.h");
    puts("- check address decoding for MTB registers");
    UartEndSimulation();  //stop simulation
    }
  else if (probing_mtb==2) {
    puts("Expected hardfault occurred.\n");
    // Print final test result
    if (Errors == 0) {
      puts("** TEST PASSED ** \n");
    } else {
      printf("** TEST FAILED ** with %d errors\n", Errors);
    }
    UartEndSimulation();  //stop simulation
    }
  else {
    //
    // The HardFault Handler is reach during the rest of the testing
    //
    // MTB Address Width mismatch could cause HardFault
    puts("Unexpected HardFault - FAIL\n");
    puts("- this could be an issue with MTB SRAM memory map setup");
    puts("  Please check :");
    puts("- EXPECTED_MTB_AWIDTH in config_id.h");
    puts("- MTB address width");
    puts("- MTB SRAM address width");
    puts("- AHB interconnect memory map (address decoder)");
    puts("  are correctly configured.");
    UartEndSimulation();
  }
}

// ----------------------------------------------------------------------------
// Error message for failures in address width detect
// ----------------------------------------------------------------------------
void fail_terminate_test(void)
{
  puts ("** ERROR **\nPlease check :");
  puts ("- EXPECTED_MTB_AWIDTH in config_id.h");
  puts ("- MTB address width");
  puts ("- MTB SRAM address width");
  puts ("- AHB interconnect memory map (address decoder)");
  puts ("  are correctly configured.");

  UartEndSimulation();  //stop simulation

}

#else
  // If not Cortex-M0+, no test code is needed
#endif
